@@ -1,17 +1,21 @@
 /*
-Copyright (C) 2000  The PARI group.
+Copyright (C) 2000-2014  The PARI group.
 
 This file is part of the GP2C package.
 
-PARI/GP is free software; you can redistribute it and/or modify it under the
-terms of the GNU General Public License as published by the Free Software
-Foundation. It is distributed in the hope that it will be useful, but WITHOUT
-ANY WARRANTY WHATSOEVER.
+GP2C is free software; you can redistribute it and/or modify it under the terms
+of the GNU General Public License as published by the Free Software Foundation;
+either version 2 of the License, or (at your option) any later version.
 
-Check the License for details. You should have received a copy of it, along
-with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+This program is distributed in the hope that it will be useful, but WITHOUT ANY
+WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
+PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
+You should have received a copy of the GNU General Public License along with
+this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
+Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
+
+#define TYPE
 #include "config.h"
 #include <stdlib.h>
 #include <stdio.h>
@@ -19,25 +23,24 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include <string.h>
 #include <errno.h>
 #include "header.h"
+
 FILE *yyin;
-int yyparse (void);
-void  yyrestart(FILE *new_file);
-void patchfunclist(void);
-void inittype(void);
-void initoperators(void);
-int linecount,errs;
+int linecount;
 int yydebug;
 static int optgen=0,opttree=0,opttype=0,optinfo=0;
 int optstrict=0;
 int optcleanvar=0;/*used in genbrace to suppress {} optimization*/
+int opthashline=0;
 int indentlevel=2;
+int do_warning=1;
 char *optprefix,*optsuffix;
-void printlistfunc(FILE *fout);
-void outputtype(FILE *fout);
-void cleancode(int n, int p);
-void cleanvar(int n);
-void initdesc(const char *s);
-void maketreeGRL(FILE *fout,int n);
+int debug;
+const char *nameparse, *namelib;
+int lastpass;
+int autogc;
+int warn;
+int safecoeff;
+
 void dump(FILE *fout)
 {
   int i,j;
@@ -51,27 +54,28 @@ void dump(FILE *fout)
       {
         if (fc->c[j].flag&(1<<Carg))
           fprintf(fout,"%s%s %s",j?", ":"",
-              Gname[fc->c[j].t],fc->c[j].var);
+              GPname(vartype(fc->c[j])),varstr(fc->c[j]));
       }
       fprintf(fout,")\n");
       if (strcmp(gp->gpname,gp->proto.cname))
         fprintf(fout,"GP name: %s\n",gp->gpname);
       fprintf(fout,"code: %s\n",gp->proto.code);
-      fprintf(fout,"return type: %s\n",Gname[gp->type]);
-      fprintf(fout,"mode=%d \t  spec=%d\n",gp->mode,gp->spec);
+      fprintf(fout,"return type: %s\n",GPname(functype(*gp)));
+      fprintf(fout,"mode=%d \t  spec=%d\n",funcmode(*gp),gp->spec);
       for(   ;j<fc->s.n;j++)
       {
         ctxvar *c=fc->c+j;
         if (!((c->flag&(1<<Carg)) || ((c->flag&(1<<Cconst)) && c->val==-1)))
-          fprintf(fout,"%s %s\n",Gname[c->t],c->var);
+          fprintf(fout,"%s %s\n",GPname(vartype(*c)),varstr(*c));
       }
       fprintf(fout,"\n");
     }
 }
 extern int indent;
-void 
+void
 init_compiler(void)
 {
+  int n;
   stack_init(&s_node,sizeof(*tree),(void **)&tree);
   stack_init(&s_value,sizeof(*value),(void **)&value);
   stack_init(&s_func,sizeof(*lfunc),(void **)&lfunc);
@@ -79,62 +83,145 @@ init_compiler(void)
   stack_init(&s_aff,sizeof(*affstack),(void **)&affstack);
   stack_init(&s_bloc,sizeof(*block),(void **)&block);
   stack_init(&s_comment,sizeof(*com),(void **)&com);
+  stack_init(&s_errors,sizeof(*errors),(void **)&errors);
+  stack_init(&s_Ctype,sizeof(*Ctype),(void **)&Ctype);
+  stack_init(&s_GPtype,sizeof(*GPtype),(void **)&GPtype);
+  stack_init(&s_Mmode,sizeof(*Mmode),(void **)&Mmode);
+  stack_init(&s_label,sizeof(*label),(void **)&label);
+   /*Node 0 should be a (Gvoid)Fnoarg*/
+  n=newnode(Fnoarg,-1,-1); tree[n].t=Gvoid;
+   /*Node 1 should be a (Gnotype)Fnoarg*/
+  n=newnode(Fnoarg,-1,-1); tree[n].t=Gnotype;
 }
+
 void
-compile(FILE *fin, char *nom)
+init_stdfunc(void)
 {
-  char *descfile;
-  int n,startnode;
+  FC_badtype=findfuncdesc("_badtype");
+  FC_formatcode=findfuncdesc("_formatcode");
+  FC_tovec=findfuncdesc("_tovec");
+  FC_tovecprec=findfuncdescopt("_tovecprec");
+  FC_cast=findfuncdesc("_cast");
+  FC_proto_ret=findfuncdesc("_proto_ret");
+  FC_proto_code=findfuncdesc("_proto_code");
+  FC_decl_base=findfuncdesc("_decl_base");
+  FC_decl_ext=findfuncdesc("_decl_ext");
+  FC_default_check=findfuncdesc("_default_check");
+  FC_default_marker=findfuncdesc("_default_marker");
+  FC_gerepileupto=findfuncdesc("_gerepileupto");
+  FC_const_smallreal=findfuncdesc("_const_smallreal");
+  FC_const_expr=findfuncdesc("_const_expr");
+  FC_copy=findfuncdesc("copy");
+  FC_avma=findfuncdesc("_avma");
+  Gpari_sp=findfunctype("_avma");
+  FC_low_stack_lim=findfuncdesc("_low_stack_lim");
+  FC_gc_needed=findfuncdescopt("_gc_needed");
+  FC_gerepileall=findfuncdesc("_gerepileall");
+  FC_forprime_init=findfuncdescopt("_forprime_init");
+  if (FC_forprime_init>=0)
+    Gforprime=strtotype("forprime");
+  else
+  {
+    Gforprime=findfunctype("_diffptr");
+    FC_forprime_next=findfuncdesc("_forprime_next");
+  }
+  Gerror=findfunctypeopt("_iferr_error");
+  FC_matrixrow=findfuncdescopt("_[_,]");
+  FC_const_real=findfuncdescopt("_const_real");
+  FC_call=findfuncdescopt("_(_)");
+  FC_strtoclosure=findfuncdescopt("_strtoclosure");
+  FC_proto_max_args=findfuncdescopt("_proto_max_args");
+  max_args = FC_proto_max_args<0 ? 8 :
+             lfunc[FC_proto_max_args].dsc->a[0].args[0].misc;
+  if (safecoeff && findfuncdescopt("_safecoeff")<0)
+  {
+    warning(-1,"gp2c option -C not supported by this PARI version");
+    safecoeff=0;
+  }
+}
+
+void
+init_typedef(void)
+{
+  gpdesc *def=lfunc[findfuncdesc("_typedef")].dsc;
+  int r;
+  ctype=(int*)calloc(s_GPtype.n,sizeof(*ctype));
+  for (r=0; r<def->nb; r++)
+  {
+    gpdescarg *rule=def->a+r;
+    int t=rule->args[0].type;
+    ctype[t]=strtoctype(rule->cname);
+  }
+  Vgen=strtoctype("GEN");
+}
+
+void
+compile(FILE *fin, FILE *fout, const char *nom)
+{
+  const char *descfile;
+  char *tmplib;
+  int startnode;
+  int status;
+  int i;
   init_compiler();
   linecount=1;
   nameparse=nom;
-  namelib=calloc(strlen(nameparse)+6,sizeof(*nameparse));
-  sprintf(namelib,"%s%s.so",nameparse[0]=='/'?"":"./",nameparse);
+  tmplib=(char*)calloc(strlen(nameparse)+6,sizeof(*nameparse));
+  sprintf(tmplib,"%s%s.so",nameparse[0]=='/'?"":"./",nameparse);
+  namelib=tmplib;
   yyin=fin;
   currfunc=-1;
   initoperators();
-  inittype();
   if (!(descfile=getenv("GP2C_FUNC_DSC")))
     descfile=FUNCDSC_PATH;
+  for(i=0;GPneeded[i];i++) newtype(GPneeded[i]);
+  for(i=0; Mneeded[i];i++) newmode( Mneeded[i]);
   initdesc(descfile);
+  inittype();
   patchfunclist();
-  if(opttype) {outputtype(stdout);exit(0);}
-
-  /*Node 0 should be a (Gvoid)Fgnil*/
-  n=newnode(Fgnil,-1,-1); tree[n].t=Gvoid;
-   /*Node 1 should be a (Gnotype)Fgnil*/
-  n=newnode(Fgnil,-1,-1); tree[n].t=Gnotype;
-  errs=0;
-  if(yyparse())
+  if(opttype) {outputtype(fout);exit(0);}
+  init_stdfunc();
+  init_typedef();
+  startnode=s_node.n-1;
+  status=yyparse();
+  for(i=0;i<s_errors.n;i++)
+  {
+    error_string *s=errors+i;
+    fprintf(stderr,"%s:%ld: %s\n",nameparse,s->lineno,s->txt);
+  }
+  if(status)
   {
     fprintf(stderr,"Errors found: aborting...\n");
     exit(1);
   }
-  if (errs)
+  if (s_errors.n)
   {
-    fprintf(stderr,"%d error%s found: aborting...\n",errs,errs==1?"":"s");
+    fprintf(stderr,"%d error%s found: aborting...\n",s_errors.n,s_errors.n==1?"":"s");
     exit(1);
   }
-  if (!s_node.n)
-  {
-    fprintf(stdout,"/*No code was generated*/\n");
-    exit(0);
-  }
-  startnode=s_node.n-1;
+  if (startnode==s_node.n-1)
+    startnode=GNIL;
+  else
+    startnode=s_node.n-1;
   if (optgen)
   {
-    printnode(stdout,startnode);
+    printnode(fout,startnode);
     printf("\n");
     return;
   }
   if (yydebug) fprintf(stderr,"End of parsing\n");
 
-  if (tree[startnode].f!=Fseq && tree[startnode].f!=Fdeffunc)
-    startnode=addseqright(startnode,GNIL);
+  startnode=addseqleft(newnode(Ffunction,newentry("_initfunc"),-1),startnode);
   gentoplevel(startnode);
   startnode=addinitfunc(startnode);
-  if (debug)
+  if (debug>1)
     printnode(stderr,startnode);
+  gentopfunc(startnode,-1,-1,-1);
+  if (debug)
+  {
+    fprintf(stderr,"\n--------------END TOPFUNC------------\n");
+    printnode(stderr,startnode);
+  }
   genblock(startnode,-1);
   if (opttree==1)
   {
@@ -143,19 +230,15 @@ compile(FILE *fin, char *nom)
   }
   if (debug)
     printnode(stderr,startnode);
-  default_type=Gnotype;
-  do
+  lastpass=0;
+  gentype(startnode);
+  do_warning=0;
+  while (lastpass)
   {
     lastpass=0;
     gentype(startnode);
-  }while(lastpass);
-  default_type=Ggen;
-  do
-  {
-    lastpass=0;
-    gentype(startnode);
-  }while(lastpass);
- 
+  }
+  do_warning=1;
   if (debug>=2)
   {
     fprintf(stderr,"\n--------------END GENTYPE------------\n");
@@ -172,18 +255,16 @@ compile(FILE *fin, char *nom)
   {
     lastpass=0;
     varlist(startnode);
-  }while(lastpass);
+  } while(lastpass);
   if (!optcleanvar)
     cleanvar(startnode);
+  pilelist(startnode);
   if (autogc)
-  {
-    pilelist(startnode);
     pileclean(startnode);
-  }
   if (optinfo) dump(stderr);
   if (!optcleanvar)
-      cleanvar(startnode);
-  cleancode(startnode,-1);
+    cleanvar(startnode);
+  cleancode(startnode,-1,left);
   if (tree[startnode].f==Fseq && tree[startnode].y==GNIL)
     startnode=tree[startnode].x;
   switch(opttree)
@@ -201,26 +282,29 @@ compile(FILE *fin, char *nom)
     printnode(stderr,startnode);
   }
   indent=0;
-  genheader(stdout);
-  gencode(stdout,startnode);
+  genheader(fout);
+  gencode(fout,startnode);
 }
 void version(void)
 {
   printf("GP to C compiler version %s \n   targeted at PARI/GP %s\n",VERSION,PARI_VERSION);
-  printf("Copyright 2000-2002 The PARI Group\n");
+  printf("Copyright 2000-2014 The PARI Group\n");
   printf("GP2C is free software, covered by the GNU General Public License, and \n\
 you are welcome to change it and/or distribute copies of it under \n\
 certain conditions.  There is absolutely no warranty for GP2C.\n");
 }
 void usage(FILE *fout, char *s)
 {
-  fprintf(fout,"%s [-ghfltvydSWTGV] [-i N] [-p <prefix>] [file.gp] \n\
+  fprintf(fout,"%s [-ghfltvydSWTGV] [-o <file>] [-i N] [-p <prefix>] [file.gp] \n\
 GP to C translator. \n\
 \n\
 user option: \n\
+ -o <file> :  Place output in file <file>. \n\
  -g : Generate automatic garbage collection code. \n\
  -iN: Set indentation level to N spaces (default 2). \n\
- -W : Output information about global variables. \n\
+ -W : Output warning about types and global variables usage. \n\
+ -C : Generate range checking code. \n\
+ -L : Generate #line directive for better C compiler messages. \n\
  -p <prefix>: Prefix user-defined symbol by <prefix> to avoid conflict. \n\
  -s <suffix>: Add suffix <suffix> to GP install names of functions. \n\
  -S : Assume strict declarations for functions. \n\
@@ -239,21 +323,23 @@ debugging options: \n\
  -V : Do not clean up variables. \n\
  \n\
 file.gp: file to be processed, default to stdin. \n\
-The generated C code is output to stdout. \n\
+The generated C code is output to stdout unless the -o option is used \n\
  \n\
-You can automated the process of compiling with \n\
-the script gp2c-run. \n\
-",s); 
+See the script gp2c-run for an automated compilation process. \n\
+",s);
 }
 int main(int argc, char **argv)
 {
-  char c;
-  FILE *fin;
+  int c;
+  FILE *fin=stdin;
+  FILE *fout=stdout;
+  const char *outfile = NULL, *infile = "stdin";
   autogc=0;
+  safecoeff=0;
   warn=0;
   optprefix=NULL;
   optsuffix=NULL;
-  while((c=getopt(argc,argv,"gi:hflp:s:StvydTGVW"))!=-1)
+  while((c=getopt(argc,argv,"gi:hflo:p:s:StvydCGLTVW"))!=-1)
   {
     switch(c)
     {
@@ -272,15 +358,18 @@ int main(int argc, char **argv)
       break;
     case 'l':
       {
-        char *descfile;
-        stack_init(&s_func,sizeof(*lfunc),(void **)&lfunc);
+        const char *descfile;
+        init_compiler();
         if (!(descfile=getenv("GP2C_FUNC_DSC")))
           descfile=FUNCDSC_PATH;
         initdesc(descfile);
         patchfunclist();
-        printlistfunc(stdout);
+        printlistfunc(fout);
         exit(0);
       }
+    case 'o':
+      outfile=strdup(optarg);
+      break;
     case 'p':
       optprefix=strdup(optarg);
       break;
@@ -298,7 +387,7 @@ int main(int argc, char **argv)
       exit(0);
     case 'd':
       debug++;
-      debug_stack=1;
+      debug_stack=debug>1;
       break;
     case 'y':
       yydebug++;
@@ -309,9 +398,15 @@ int main(int argc, char **argv)
     case 'G':
       optgen=1-optgen;
       break;
+    case 'L':
+      opthashline=1-opthashline;
+      break;
     case 'V':
       optcleanvar=1-optcleanvar;
       break;
+    case 'C':
+      safecoeff=1-safecoeff;
+      break;
     case 'W':
       warn=1-warn;
       break;
@@ -326,19 +421,22 @@ int main(int argc, char **argv)
     usage(stderr,argv[0]);
     exit(1);
   }
-  if (argc==optind)
-    compile(stdin,"stdin");
-  else
+  if (argc!=optind)
   {
-    char *nom=argv[optind];
-    char *nomf=nom;
-    if ((fin=fopen(nomf,"r"))==NULL)
+    infile = argv[optind];
+    if ((fin=fopen(infile,"r"))==NULL)
     {
       perror(argv[0]);
       exit(errno);
     }
-    compile(fin,nom);
-    fclose(fin);
   }
+  if (outfile && ((fout=fopen(outfile,"w"))==NULL))
+  {
+    perror(argv[0]);
+    exit(errno);
+  }
+  compile(fin,fout,infile);
+  if (fin!=stdin) fclose(fin);
+  if (fout!=stdout) fclose(fout);
   return 0;
 }
